#include <stdio.h>
#include <stdlib.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include "ido.h"


#define   TAU   8.0
#define   US(x, y, t) ( - 2.0*cos(M_PI*(t)/TAU)*sin(M_PI*(x))*sin(M_PI*(x))*cos(M_PI*(y))*sin(M_PI*(y)) )
#define   VS(x, y, t) ( 2.0*cos(M_PI*(t)/TAU)*cos(M_PI*(x))*sin(M_PI*(x))*sin(M_PI*(y))*sin(M_PI*(y)) )



__global__  void  gpu_ppush1
// ====================================================================
//
// program    :  CUDA device code for the particle push calculation
//               by 4-stage Runge-Kutta time integration
//
// date       :  Mar 02, 2009
// programmer :  Takayuki Aoki
// place      :  Tokyo Institute of Technology
//
(
   float    *x,         /* x-coordinate of the particles             */
   float    *y,         /* y-coordinate of the particles             */
   float    *xn,        /* updated x-coordinate of the particles     */
   float    *yn,        /* updated y-coordinate of the particles     */
   float    time,       /* time                                      */
   float    dt          /* time step interval                        */
)
// --------------------------------------------------------------------
{
   int   j,   jx,  jy;
   float  xg,  yg,  xtdt,  ytdt;

       jx = blockDim.x*blockIdx.x + threadIdx.x;
       jy = blockDim.y*blockIdx.y;
       j = gridDim.x*blockDim.x*jy + jx;

       xg = x[j];    yg = y[j];

       xtdt = US(xg, yg, time)*dt;          ytdt = VS(xg, yg, time)*dt;
       xn[j] = xg + xtdt;                   yn[j] = yg + ytdt;
}



__global__  void  gpu_ppush4
// ====================================================================
//
// program    :  CUDA device code for the particle push calculation
//               by 4-stage Runge-Kutta time integration
//
// date       :  Mar 02, 2009
// programmer :  Takayuki Aoki
// place      :  Tokyo Institute of Technology
//
(
   float    *x,         /* x-coordinate of the particles             */
   float    *y,         /* y-coordinate of the particles             */
   float    *xn,        /* updated x-coordinate of the particles     */
   float    *yn,        /* updated y-coordinate of the particles     */
   float    time,       /* time                                      */
   float    dt          /* time step interval                        */
)
// --------------------------------------------------------------------
{
   int   j,   jx,  jy;
   float  xg,  yg,  xtdt,  ytdt,  xm,  ym,  xs,  ys;

       jx = blockDim.x*blockIdx.x + threadIdx.x;
       jy = blockDim.y*blockIdx.y;
       j = gridDim.x*blockDim.x*jy + jx;

       xg = x[j];    yg = y[j];

       xtdt = US(xg, yg, time)*dt;          ytdt = VS(xg, yg, time)*dt;
       xm = xg + 0.5*xtdt;                  ym = yg + 0.5*ytdt;
       xs = xg + xtdt/6.0;                  ys = yg + ytdt/6.0;

       xtdt = US(xm, ym, time+0.5*dt)*dt;   ytdt = VS(xm, ym, time+0.5*dt)*dt;
       xm = xg + 0.5*xtdt;                  ym = yg + 0.5*ytdt;
       xs += xtdt/3.0;                      ys += ytdt/3.0;

       xtdt = US(xm, ym, time+0.5*dt)*dt;   ytdt = VS(xm, ym, time+0.5*dt)*dt;
       xm = xg + xtdt;                      ym = yg + ytdt;
       xs += xtdt/3.0;                      ys += ytdt/3.0;

       xtdt = US(xm, ym, time+dt)*dt;       ytdt = VS(xm, ym, time+dt)*dt;
       xn[j] = xs + xtdt/6.0;               yn[j] = ys + ytdt/6.0;
}



void   cpu_ppush4
// ====================================================================
//
// purpos     :  Particle push by 4-stage Runge-Kutta time integration
//
// date       :  May 12, 2009
// programmer :  Takayuki Aoki
// place      :  Tokyo Institute of Technology
//
(
   int      np,         /* number of the particles                   */
   float    *x,         /* x-coordinate of the particles             */
   float    *y,         /* y-coordinate of the particles             */
   float    *xn,        /* updated x-coordinate of the particles     */
   float    *yn,        /* updated y-coordinate of the particles     */
   float    time,       /* time                                      */
   float    dt          /* time step interval                        */
)
// --------------------------------------------------------------------
{
   int   j;
   float  xt,  xm,  yt,  ym;

   for(j = 0; j < np; j++) {
       xt = US(x[j],y[j],time);        yt = VS(x[j],y[j],time);
       xm = x[j] + 0.5*xt*dt;          ym = y[j] + 0.5*yt*dt;
       xn[j] = x[j] + xt*dt/6.0;       yn[j] = y[j] + yt*dt/6.0;

       xt = US(xm, ym, time+0.5*dt);   yt = VS(xm, ym, time+0.5*dt);
       xm = x[j] + 0.5*xt*dt;          ym = y[j] + 0.5*yt*dt;
       xn[j] += xt*dt/3.0;             yn[j] += yt*dt/3.0;

       xt = US(xm, ym, time+0.5*dt);   yt = VS(xm, ym, time+0.5*dt);
       xm = x[j] + xt*dt;              ym = y[j] + yt*dt;
       xn[j] += xt*dt/3.0;             yn[j] += yt*dt/3.0;

       xt = US(xm, ym, time+dt);       yt = VS(xm, ym, time+dt);
       xn[j] += xt*dt/6.0;             yn[j] += yt*dt/6.0;
   }
}



void  cpu_ppush1
// ====================================================================
//
// purpos     :  Particle push by 4-stage Runge-Kutta time integration
//
// date       :  May 12, 2009
// programmer :  Takayuki Aoki
// place      :  Tokyo Institute of Technology
//
(
   int      np,         /* number of the particles                   */
   float    *x,         /* x-coordinate of the particles             */
   float    *y,         /* y-coordinate of the particles             */
   float    *xn,        /* updated x-coordinate of the particles     */
   float    *yn,        /* updated y-coordinate of the particles     */
   float    time,       /* time                                      */
   float    dt          /* time step interval                        */
)
// --------------------------------------------------------------------
{
   int     j;
   float   xt,  yt;

   for(j = 0; j < np; j++) {
       xt = US(x[j], y[j], time);       yt = VS(x[j], y[j], time);
       xn[j] = x[j] + xt*dt;            yn[j] = y[j] + yt*dt;
   }
}



void  ppush
// ====================================================================
//
// purpos     :  Particle push by 4-stage Runge-Kutta time integration
//
// date       :  May 12, 2009
// programmer :  Takayuki Aoki
// place      :  Tokyo Institute of Technology
//
(
   int      numGPUs,    /* available number of GPU device            */
   int      np,         /* number of the particles                   */
   float    *x,         /* x-coordinate of the particles             */
   float    *y,         /* y-coordinate of the particles             */
   float    *xn,        /* updated x-coordinate of the particles     */
   float    *yn,        /* updated y-coordinate of the particles     */
   float    time,       /* time                                      */
   float    dt          /* time step interval                        */
)
// --------------------------------------------------------------------
{
     if(numGPUs > 0) {

        dim3  Dg(np/256,1,1),  Db(256,1,1);
#ifdef RUNGE_KUTTA4
        gpu_ppush4<<< Dg, Db >>>(x,y,xn,yn,time,dt);
#else
        gpu_ppush1<<< Dg, Db >>>(x,y,xn,yn,time,dt);
#endif
        cudaThreadSynchronize();

     }
     else {
#ifdef RUNGE_KUTTA4
        cpu_ppush4(np,x,y,xn,yn,time,dt);
#else
        cpu_ppush1(np,x,y,xn,yn,time,dt);
#endif
     }
}
